#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING

from argilla_server.constants import DEFAULT_API_KEY, DEFAULT_PASSWORD, DEFAULT_USERNAME
from argilla_server.contexts import accounts
from argilla_server.models import User, UserRole

if TYPE_CHECKING:
    from sqlalchemy.orm import Session
    from typer import Typer
    from typer.testing import CliRunner


def test_create_default(sync_db: "Session", cli_runner: "CliRunner", cli: "Typer"):
    result = cli_runner.invoke(cli, "database users create_default")

    assert result.exit_code == 0
    assert result.output != ""
    assert sync_db.query(User).count() == 1

    default_user = sync_db.query(User).filter_by(username=DEFAULT_USERNAME).first()
    assert default_user
    assert default_user.role == UserRole.owner
    assert default_user.api_key == DEFAULT_API_KEY
    assert accounts.verify_password(DEFAULT_PASSWORD, default_user.password_hash)
    assert [ws.name for ws in default_user.workspaces] == [DEFAULT_USERNAME]


def test_create_default_with_specific_api_key_and_password(sync_db: "Session", cli_runner: "CliRunner", cli: "Typer"):
    result = cli_runner.invoke(cli, "database users create_default --api-key my-api-key --password my-password")

    assert result.exit_code == 0
    assert result.output != ""
    assert sync_db.query(User).count() == 1

    default_user = sync_db.query(User).filter_by(username=DEFAULT_USERNAME).first()
    assert default_user
    assert default_user.role == UserRole.owner
    assert default_user.api_key == "my-api-key"
    assert accounts.verify_password("my-password", default_user.password_hash)
    assert [ws.name for ws in default_user.workspaces] == [DEFAULT_USERNAME]


def test_create_default_quiet(sync_db: "Session", cli_runner: "CliRunner", cli: "Typer"):
    result = cli_runner.invoke(cli, "database users create_default --quiet")

    assert result.exit_code == 0
    assert result.output == ""
    assert sync_db.query(User).count() == 1


def test_create_default_with_existent_default_user(sync_db: "Session", cli_runner: "CliRunner", cli: "Typer"):
    result = cli_runner.invoke(cli, "database users create_default")

    assert result.exit_code == 0
    assert result.output != ""
    assert sync_db.query(User).count() == 1

    result = cli_runner.invoke(cli, "database users create_default")

    assert result.exit_code == 0
    assert result.output == "User with default username already found on database, will not do anything.\n"
    assert sync_db.query(User).count() == 1


def test_create_default_with_existent_default_user_and_quiet(sync_db: "Session", cli_runner: "CliRunner", cli: "Typer"):
    result = cli_runner.invoke(cli, "database users create_default")

    assert result.exit_code == 0
    assert result.output != ""
    assert sync_db.query(User).count() == 1

    result = cli_runner.invoke(cli, "database users create_default --quiet")

    assert result.exit_code == 0
    assert result.output == ""
    assert sync_db.query(User).count() == 1



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla_server.commons.models import TaskStatus
from argilla_server.daos.backend.metrics.base import (
    HistogramAggregation,
    MetadataAggregations,
    TermsAggregation,
    WordCloudAggregation,
)

METRICS = {
    "text_length": HistogramAggregation(
        id="text_length",
        field="metrics.text_length",
        script="params._source.text.length()",
        fixed_interval=1,
    ),
    "error_distribution": TermsAggregation(
        id="error_distribution",
        field="predicted",
        missing="unknown",
        fixed_size=3,
    ),
    "status_distribution": TermsAggregation(
        id="status_distribution",
        field="status",
        fixed_size=len(TaskStatus),
    ),
    "words_cloud": WordCloudAggregation(
        id="words_cloud",
        default_field="text.wordcloud",
    ),
    "metadata": MetadataAggregations(
        id="metadata",
    ),
    "predicted_by": TermsAggregation(
        id="predicted_by",
        field="predicted_by",
    ),
    "annotated_by": TermsAggregation(
        id="annotated_by",
        field="annotated_by",
    ),
    "score": HistogramAggregation(
        id="score",
        field="score",
        fixed_interval=0.001,
    ),
}



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Depends

from argilla_server.services.info import ApiInfo, ApiInfoService, ApiStatus

router = APIRouter(tags=["status"])


@router.get("/_status", operation_id="api_status", response_model=ApiStatus)
async def api_status(service: ApiInfoService = Depends(ApiInfoService.get_instance)) -> ApiStatus:
    return service.api_status()


@router.get("/_info", operation_id="api_info", response_model=ApiInfo)
async def api_info(service: ApiInfoService = Depends(ApiInfoService.get_instance)) -> ApiInfo:
    return ApiInfo.parse_obj(service.api_status())



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from unittest import mock

from argilla_server.security.settings import Settings


def test_default_security_settings():
    settings = Settings()

    assert settings.algorithm == "HS256"
    assert settings.token_expiration == 86400
    assert settings.oauth_cfg == ".oauth.yaml"

    assert len(settings.secret_key) == 32


def test_configure_algorithm():
    algorithm = "mock-algorithm"
    with mock.patch.dict(os.environ, {"ARGILLA_AUTH_ALGORITHM": algorithm}):
        settings = Settings()

        assert settings.algorithm == algorithm


def test_configure_token_expiration():
    token_expiration = 3600
    with mock.patch.dict(os.environ, {"ARGILLA_AUTH_TOKEN_EXPIRATION": str(token_expiration)}):
        settings = Settings()

        assert settings.token_expiration == token_expiration


def test_configure_secret_key():
    secret_key = "mock-secret-key"
    with mock.patch.dict(os.environ, {"ARGILLA_AUTH_SECRET_KEY": secret_key}):
        settings = Settings()

        assert settings.secret_key == secret_key


def test_configure_oauth_cfg():
    oauth_cfg = "mock-oauth-cfg"
    with mock.patch.dict(os.environ, {"ARGILLA_AUTH_OAUTH_CFG": oauth_cfg}):
        settings = Settings()

        assert settings.oauth_cfg == oauth_cfg


def test_configure_algorithm_with_old_env_name():
    algorithm = "old_mock-algorithm"
    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_ALGORITHM": algorithm}):
        settings = Settings()

        assert settings.algorithm == algorithm


# TODO: remove this test when ARGILLA_LOCAL_AUTH_SECRET_KEY will be removed
def test_configure_secret_key_with_old_env_name():
    secret_key = "old_mock-secret-key"
    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_SECRET_KEY": secret_key}):
        settings = Settings()

        assert settings.secret_key == secret_key


# TODO: remove this test when ARGILLA_LOCAL_AUTH_TOKEN_EXPIRATION_IN_MINUTES will be removed
def test_configure_token_expiration_in_minutes():
    token_expiration_in_minutes = 30
    with mock.patch.dict(
        os.environ, {"ARGILLA_LOCAL_AUTH_TOKEN_EXPIRATION_IN_MINUTES": str(token_expiration_in_minutes)}
    ):
        settings = Settings()

        assert settings.token_expiration == token_expiration_in_minutes * 60



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.models import MetadataProperty, User
from argilla_server.policies import MetadataPropertyPolicyV1, authorize
from argilla_server.schemas.v1.metadata_properties import MetadataMetrics, MetadataProperty, MetadataPropertyUpdate
from argilla_server.search_engine import SearchEngine, get_search_engine
from argilla_server.security import auth

router = APIRouter(tags=["metadata properties"])


async def _get_metadata_property(db: "AsyncSession", metadata_property_id: UUID) -> "MetadataProperty":
    metadata_property = await datasets.get_metadata_property_by_id(db, metadata_property_id)
    if not metadata_property:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Metadata property with id `{metadata_property_id}` not found",
        )

    return metadata_property


@router.get("/metadata-properties/{metadata_property_id}/metrics", response_model=MetadataMetrics)
async def get_metadata_property_metrics(
    *,
    db: AsyncSession = Depends(get_async_db),
    metadata_property_id: UUID,
    search_engine: SearchEngine = Depends(get_search_engine),
    current_user: User = Security(auth.get_current_user),
):
    metadata_property = await _get_metadata_property(db, metadata_property_id)

    await authorize(current_user, MetadataPropertyPolicyV1.get(metadata_property))

    return await search_engine.compute_metrics_for(metadata_property)


@router.patch("/metadata-properties/{metadata_property_id}", response_model=MetadataProperty)
async def update_metadata_property(
    *,
    db: AsyncSession = Depends(get_async_db),
    metadata_property_id: UUID,
    metadata_property_update: MetadataPropertyUpdate,
    current_user: User = Security(auth.get_current_user),
):
    metadata_property = await _get_metadata_property(db, metadata_property_id)

    await authorize(current_user, MetadataPropertyPolicyV1.update(metadata_property))

    return await datasets.update_metadata_property(db, metadata_property, metadata_property_update)


@router.delete("/metadata-properties/{metadata_property_id}", response_model=MetadataProperty)
async def delete_metadata_property(
    *,
    db: AsyncSession = Depends(get_async_db),
    metadata_property_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    metadata_property = await _get_metadata_property(db, metadata_property_id)

    await authorize(current_user, MetadataPropertyPolicyV1.delete(metadata_property))

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        await datasets.delete_metadata_property(db, metadata_property)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))

    return metadata_property



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import secrets
from datetime import datetime
from typing import Any, List, Optional, Union
from uuid import UUID

from sqlalchemy import JSON, ForeignKey, String, Text, UniqueConstraint, and_, sql
from sqlalchemy import Enum as SAEnum
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import Mapped, mapped_column, relationship

from argilla_server.enums import (
    DatasetStatus,
    MetadataPropertyType,
    QuestionType,
    ResponseStatus,
    SuggestionType,
    UserRole,
)
from argilla_server.models.base import DatabaseModel
from argilla_server.models.metadata_properties import MetadataPropertySettings
from argilla_server.models.mixins import inserted_at_current_value
from argilla_server.models.questions import QuestionSettings
from argilla_server.pydantic_v1 import parse_obj_as

# Include here the data model ref to be accessible for automatic alembic migration scripts
__all__ = [
    "Dataset",
    "Field",
    "Question",
    "Record",
    "Response",
    "Suggestion",
    "User",
    "Workspace",
    "WorkspaceUser",
    "MetadataProperty",
    "Vector",
    "VectorSettings",
]

_USER_API_KEY_BYTES_LENGTH = 80


class Field(DatabaseModel):
    __tablename__ = "fields"

    name: Mapped[str] = mapped_column(String, index=True)
    title: Mapped[str] = mapped_column(Text)
    required: Mapped[bool] = mapped_column(default=False)
    settings: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSON), default={})
    dataset_id: Mapped[UUID] = mapped_column(ForeignKey("datasets.id", ondelete="CASCADE"), index=True)

    dataset: Mapped["Dataset"] = relationship(back_populates="fields")

    __table_args__ = (UniqueConstraint("name", "dataset_id", name="field_name_dataset_id_uq"),)

    def __repr__(self):
        return (
            f"Field(id={str(self.id)!r}, name={self.name!r}, required={self.required!r}, "
            f"dataset_id={str(self.dataset_id)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


ResponseStatusEnum = SAEnum(ResponseStatus, name="response_status_enum")


class Response(DatabaseModel):
    __tablename__ = "responses"

    values: Mapped[Optional[dict]] = mapped_column(MutableDict.as_mutable(JSON))
    status: Mapped[ResponseStatus] = mapped_column(ResponseStatusEnum, default=ResponseStatus.submitted, index=True)
    record_id: Mapped[UUID] = mapped_column(ForeignKey("records.id", ondelete="CASCADE"), index=True)
    user_id: Mapped[Optional[UUID]] = mapped_column(ForeignKey("users.id", ondelete="SET NULL"), index=True)

    record: Mapped["Record"] = relationship(back_populates="responses")
    user: Mapped["User"] = relationship(back_populates="responses")

    __table_args__ = (UniqueConstraint("record_id", "user_id", name="response_record_id_user_id_uq"),)
    __upsertable_columns__ = {"values", "status"}

    @property
    def is_submitted(self):
        return self.status == ResponseStatus.submitted

    def __repr__(self):
        return (
            f"Response(id={str(self.id)!r}, record_id={str(self.record_id)!r}, user_id={str(self.user_id)!r}, "
            f"status={self.status.value!r}, inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


SuggestionTypeEnum = SAEnum(SuggestionType, name="suggestion_type_enum")


class Suggestion(DatabaseModel):
    __tablename__ = "suggestions"

    value: Mapped[Any] = mapped_column(JSON)
    score: Mapped[Optional[float]] = mapped_column(nullable=True)
    agent: Mapped[Optional[str]] = mapped_column(nullable=True)
    type: Mapped[Optional[SuggestionType]] = mapped_column(SuggestionTypeEnum, nullable=True, index=True)
    record_id: Mapped[UUID] = mapped_column(ForeignKey("records.id", ondelete="CASCADE"), index=True)
    question_id: Mapped[UUID] = mapped_column(ForeignKey("questions.id", ondelete="CASCADE"), index=True)

    record: Mapped["Record"] = relationship(back_populates="suggestions")
    question: Mapped["Question"] = relationship(back_populates="suggestions")

    __table_args__ = (UniqueConstraint("record_id", "question_id", name="suggestion_record_id_question_id_uq"),)
    __upsertable_columns__ = {"value", "score", "agent", "type"}

    def __repr__(self) -> str:
        return (
            f"Suggestion(id={self.id}, score={self.score}, agent={self.agent}, type={self.type}, "
            f"record_id={self.record_id}, question_id={self.question_id}, inserted_at={self.inserted_at}, "
            f"updated_at={self.updated_at})"
        )


class Vector(DatabaseModel):
    __tablename__ = "vectors"

    value: Mapped[List[Any]] = mapped_column(JSON)
    record_id: Mapped[UUID] = mapped_column(ForeignKey("records.id", ondelete="CASCADE"), index=True)
    vector_settings_id: Mapped[UUID] = mapped_column(ForeignKey("vectors_settings.id", ondelete="CASCADE"), index=True)

    record: Mapped["Record"] = relationship(back_populates="vectors")
    vector_settings: Mapped["VectorSettings"] = relationship(back_populates="vectors")

    __table_args__ = (
        UniqueConstraint("record_id", "vector_settings_id", name="vector_record_id_vector_settings_id_uq"),
    )
    __upsertable_columns__ = {"value"}

    def __repr__(self) -> str:
        return (
            f"Vector(id={self.id}, vector_settings_id={self.vector_settings_id}, record_id={self.record_id}, "
            f"inserted_at={self.inserted_at}, updated_at={self.updated_at})"
        )


class VectorSettings(DatabaseModel):
    __tablename__ = "vectors_settings"

    name: Mapped[str] = mapped_column(index=True)
    title: Mapped[str] = mapped_column(Text)
    dimensions: Mapped[int] = mapped_column()
    dataset_id: Mapped[UUID] = mapped_column(ForeignKey("datasets.id", ondelete="CASCADE"), index=True)

    dataset: Mapped["Dataset"] = relationship(back_populates="vectors_settings")
    vectors: Mapped[List["Vector"]] = relationship(
        back_populates="vector_settings",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Vector.inserted_at.asc(),
    )

    __table_args__ = (UniqueConstraint("name", "dataset_id", name="vector_settings_name_dataset_id_uq"),)

    def __repr__(self) -> str:
        return (
            f"VectorSettings(id={self.id}, name={self.name}, dimensions={self.dimensions}, "
            f"dataset_id={self.dataset_id}, inserted_at={self.inserted_at}, updated_at={self.updated_at})"
        )


class Record(DatabaseModel):
    __tablename__ = "records"

    fields: Mapped[dict] = mapped_column(JSON, default={})
    metadata_: Mapped[Optional[dict]] = mapped_column("metadata", MutableDict.as_mutable(JSON), nullable=True)
    external_id: Mapped[Optional[str]] = mapped_column(index=True)
    dataset_id: Mapped[UUID] = mapped_column(ForeignKey("datasets.id", ondelete="CASCADE"), index=True)

    dataset: Mapped["Dataset"] = relationship(back_populates="records")
    responses: Mapped[List["Response"]] = relationship(
        back_populates="record",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Response.inserted_at.asc(),
    )
    suggestions: Mapped[List["Suggestion"]] = relationship(
        back_populates="record",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Suggestion.inserted_at.asc(),
    )
    vectors: Mapped[List["Vector"]] = relationship(
        back_populates="record",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Vector.inserted_at.asc(),
    )

    __table_args__ = (UniqueConstraint("external_id", "dataset_id", name="record_external_id_dataset_id_uq"),)

    def __repr__(self):
        return (
            f"Record(id={str(self.id)!r}, external_id={self.external_id!r}, dataset_id={str(self.dataset_id)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )

    def vector_value_by_vector_settings(self, vector_settings: "VectorSettings") -> Union[List[float], None]:
        for vector in self.vectors:
            if vector.vector_settings_id == vector_settings.id:
                return vector.value


class Question(DatabaseModel):
    __tablename__ = "questions"

    name: Mapped[str] = mapped_column(String, index=True)
    title: Mapped[str] = mapped_column(Text)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    required: Mapped[bool] = mapped_column(default=False)
    settings: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSON), default={})
    dataset_id: Mapped[UUID] = mapped_column(ForeignKey("datasets.id", ondelete="CASCADE"), index=True)

    dataset: Mapped["Dataset"] = relationship(back_populates="questions")
    suggestions: Mapped[List["Suggestion"]] = relationship(
        back_populates="question",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Suggestion.inserted_at.asc(),
    )

    __table_args__ = (UniqueConstraint("name", "dataset_id", name="question_name_dataset_id_uq"),)

    @property
    def parsed_settings(self) -> QuestionSettings:
        return parse_obj_as(QuestionSettings, self.settings)

    @property
    def type(self) -> QuestionType:
        return QuestionType(self.settings["type"])

    def __repr__(self):
        return (
            f"Question(id={str(self.id)!r}, name={self.name!r}, required={self.required!r}, "
            f"dataset_id={str(self.dataset_id)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


class MetadataProperty(DatabaseModel):
    __tablename__ = "metadata_properties"

    name: Mapped[str] = mapped_column(String, index=True)
    title: Mapped[str] = mapped_column(Text)
    settings: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSON), default={})
    allowed_roles: Mapped[List[UserRole]] = mapped_column(MutableList.as_mutable(JSON), default=[], server_default="[]")
    dataset_id: Mapped[UUID] = mapped_column(ForeignKey("datasets.id", ondelete="CASCADE"), index=True)

    dataset: Mapped["Dataset"] = relationship(back_populates="metadata_properties")

    __table_args__ = (UniqueConstraint("name", "dataset_id", name="metadata_property_name_dataset_id_uq"),)

    @property
    def type(self) -> MetadataPropertyType:
        return MetadataPropertyType(self.settings["type"])

    @property
    def parsed_settings(self) -> MetadataPropertySettings:
        return parse_obj_as(MetadataPropertySettings, self.settings)

    @property
    def visible_for_annotators(self) -> bool:
        return UserRole.annotator in self.allowed_roles

    def __repr__(self):
        return (
            f"MetadataProperty(id={str(self.id)!r}, name={self.name!r}, dataset_id={str(self.dataset_id)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


DatasetStatusEnum = SAEnum(DatasetStatus, name="dataset_status_enum")


def _updated_at_current_value(context: DefaultExecutionContext) -> datetime:
    return context.get_current_parameters(isolate_multiinsert_groups=False)["updated_at"]


class Dataset(DatabaseModel):
    __tablename__ = "datasets"

    name: Mapped[str] = mapped_column(index=True)
    guidelines: Mapped[Optional[str]] = mapped_column(Text)
    allow_extra_metadata: Mapped[bool] = mapped_column(default=True, server_default=sql.true())
    status: Mapped[DatasetStatus] = mapped_column(DatasetStatusEnum, default=DatasetStatus.draft, index=True)
    workspace_id: Mapped[UUID] = mapped_column(ForeignKey("workspaces.id", ondelete="CASCADE"), index=True)
    inserted_at: Mapped[datetime] = mapped_column(default=datetime.utcnow)
    updated_at: Mapped[datetime] = mapped_column(default=inserted_at_current_value, onupdate=datetime.utcnow)
    last_activity_at: Mapped[datetime] = mapped_column(
        default=inserted_at_current_value, onupdate=_updated_at_current_value
    )

    workspace: Mapped["Workspace"] = relationship(back_populates="datasets")
    fields: Mapped[List["Field"]] = relationship(
        back_populates="dataset",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Field.inserted_at.asc(),
    )
    questions: Mapped[List["Question"]] = relationship(
        back_populates="dataset",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Question.inserted_at.asc(),
    )
    records: Mapped[List["Record"]] = relationship(
        back_populates="dataset",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=Record.inserted_at.asc(),
    )
    metadata_properties: Mapped[List["MetadataProperty"]] = relationship(
        back_populates="dataset",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=MetadataProperty.inserted_at.asc(),
    )
    vectors_settings: Mapped[List["VectorSettings"]] = relationship(
        back_populates="dataset",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=VectorSettings.inserted_at.asc(),
    )

    __table_args__ = (UniqueConstraint("name", "workspace_id", name="dataset_name_workspace_id_uq"),)

    @property
    def is_draft(self):
        return self.status == DatasetStatus.draft

    @property
    def is_ready(self):
        return self.status == DatasetStatus.ready

    def __repr__(self):
        return (
            f"Dataset(id={str(self.id)!r}, name={self.name!r}, guidelines={self.guidelines!r}, "
            f"status={self.status.value!r}, workspace_id={str(self.workspace_id)!r}, "
            f"last_activity_at={str(self.last_activity_at)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


class WorkspaceUser(DatabaseModel):
    __tablename__ = "workspaces_users"

    workspace_id: Mapped[UUID] = mapped_column(ForeignKey("workspaces.id", ondelete="CASCADE"), index=True)
    user_id: Mapped[UUID] = mapped_column(ForeignKey("users.id", ondelete="CASCADE"), index=True)

    workspace: Mapped["Workspace"] = relationship(viewonly=True)
    user: Mapped["User"] = relationship(viewonly=True)

    __table_args__ = (UniqueConstraint("workspace_id", "user_id", name="workspace_id_user_id_uq"),)

    def __repr__(self):
        return (
            f"WorkspaceUser(id={str(self.id)!r}, workspace_id={str(self.workspace_id)!r}, "
            f"user_id={str(self.user_id)!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


class Workspace(DatabaseModel):
    __tablename__ = "workspaces"

    name: Mapped[str] = mapped_column(unique=True, index=True)

    datasets: Mapped[List["Dataset"]] = relationship(back_populates="workspace", order_by=Dataset.inserted_at.asc())
    users: Mapped[List["User"]] = relationship(
        secondary="workspaces_users", back_populates="workspaces", order_by=WorkspaceUser.inserted_at.asc()
    )

    def __repr__(self):
        return (
            f"Workspace(id={str(self.id)!r}, name={self.name!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )


def generate_user_api_key() -> str:
    return secrets.token_urlsafe(_USER_API_KEY_BYTES_LENGTH)


UserRoleEnum = SAEnum(UserRole, name="user_role_enum")


class User(DatabaseModel):
    __tablename__ = "users"

    first_name: Mapped[str]
    last_name: Mapped[Optional[str]]
    username: Mapped[str] = mapped_column(unique=True, index=True)
    role: Mapped[UserRole] = mapped_column(UserRoleEnum, default=UserRole.annotator, index=True)
    api_key: Mapped[str] = mapped_column(Text, unique=True, index=True, default=generate_user_api_key)
    password_hash: Mapped[str] = mapped_column(Text)

    workspaces: Mapped[List["Workspace"]] = relationship(
        secondary="workspaces_users", back_populates="users", order_by=WorkspaceUser.inserted_at.asc()
    )
    responses: Mapped[List["Response"]] = relationship(back_populates="user")
    datasets: Mapped[List["Dataset"]] = relationship(
        secondary="workspaces_users",
        primaryjoin="User.id == WorkspaceUser.user_id",
        secondaryjoin=and_(
            Workspace.id == Dataset.workspace_id,
            WorkspaceUser.workspace_id == Workspace.id,
        ),
        viewonly=True,
        order_by=Dataset.inserted_at.asc(),
    )

    @property
    def is_owner(self):
        return self.role == UserRole.owner

    @property
    def is_admin(self):
        return self.role == UserRole.admin

    @property
    def is_annotator(self):
        return self.role == UserRole.annotator

    def __repr__(self):
        return (
            f"User(id={str(self.id)!r}, first_name={self.first_name!r}, last_name={self.last_name!r}, "
            f"username={self.username!r}, role={self.role.value!r}, "
            f"inserted_at={str(self.inserted_at)!r}, updated_at={str(self.updated_at)!r})"
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from fastapi import Body
from typing_extensions import Annotated

from argilla_server.models import ResponseStatus
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.schemas.v1.questions import QuestionName

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

RESPONSES_BULK_CREATE_MIN_ITEMS = 1
RESPONSES_BULK_CREATE_MAX_ITEMS = 100


class ResponseValue(BaseModel):
    value: Any


class ResponseValueCreate(BaseModel):
    value: Any


class ResponseValueUpdate(BaseModel):
    value: Any


class Response(BaseModel):
    id: UUID
    values: Optional[Dict[str, ResponseValue]]
    status: ResponseStatus
    record_id: UUID
    user_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class ResponseCreate(BaseModel):
    values: Optional[Dict[str, ResponseValueCreate]]
    status: ResponseStatus


class ResponseFilterScope(BaseModel):
    entity: Literal["response"]
    question: Optional[QuestionName]
    property: Optional[Literal["status"]]


class SubmittedResponseUpdate(BaseModel):
    values: Dict[str, ResponseValueUpdate]
    status: Literal[ResponseStatus.submitted]


class DiscardedResponseUpdate(BaseModel):
    values: Optional[Dict[str, ResponseValueUpdate]]
    status: Literal[ResponseStatus.discarded]


class DraftResponseUpdate(BaseModel):
    values: Optional[Dict[str, ResponseValueUpdate]]
    status: Literal[ResponseStatus.draft]


ResponseUpdate = Annotated[
    Union[SubmittedResponseUpdate, DiscardedResponseUpdate, DraftResponseUpdate],
    Body(..., discriminator="status"),
]


class SubmittedResponseUpsert(BaseModel):
    values: Dict[str, ResponseValueUpdate]
    status: Literal[ResponseStatus.submitted]
    record_id: UUID


class DiscardedResponseUpsert(BaseModel):
    values: Optional[Dict[str, ResponseValueUpdate]]
    status: Literal[ResponseStatus.discarded]
    record_id: UUID


class DraftResponseUpsert(BaseModel):
    values: Optional[Dict[str, ResponseValueUpdate]]
    status: Literal[ResponseStatus.draft]
    record_id: UUID


ResponseUpsert = Annotated[
    Union[SubmittedResponseUpsert, DiscardedResponseUpsert, DraftResponseUpsert],
    Body(..., discriminator="status"),
]


class ResponsesBulkCreate(BaseModel):
    items: List[ResponseUpsert] = Field(
        ...,
        min_items=RESPONSES_BULK_CREATE_MIN_ITEMS,
        max_items=RESPONSES_BULK_CREATE_MAX_ITEMS,
    )


class ResponseBulkError(BaseModel):
    detail: str


class ResponseBulk(BaseModel):
    item: Optional[Response]
    error: Optional[ResponseBulkError]


class ResponsesBulk(BaseModel):
    items: List[ResponseBulk]


class UserDraftResponseCreate(BaseModel):
    user_id: UUID
    values: Dict[str, ResponseValueCreate]
    status: Literal[ResponseStatus.draft]


class UserDiscardedResponseCreate(BaseModel):
    user_id: UUID
    values: Optional[Dict[str, ResponseValueCreate]]
    status: Literal[ResponseStatus.discarded]


class UserSubmittedResponseCreate(BaseModel):
    user_id: UUID
    values: Dict[str, ResponseValueCreate]
    status: Literal[ResponseStatus.submitted]


UserResponseCreate = Annotated[
    Union[UserSubmittedResponseCreate, UserDraftResponseCreate, UserDiscardedResponseCreate],
    Field(discriminator="status"),
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.apis.v0.models.commons.model import ScoreRange
from argilla_server.apis.v0.models.text_classification import TextClassificationQuery
from argilla_server.daos.backend.search.model import (
    SortableField,
    SortConfig,
    SortOrder,
)
from argilla_server.daos.backend.search.query_builder import EsQueryBuilder


@pytest.mark.parametrize(
    ["index_schema", "sort_cfg", "expected_sort"],
    [
        (
            {
                "mappings": {
                    "properties": {
                        "id": {"type": "text"},
                    }
                }
            },
            [SortableField(id="id")],
            [{"id.keyword": {"order": SortOrder.asc}}],
        ),
        (
            {
                "mappings": {
                    "properties": {
                        "id": {"type": "keyword"},
                    }
                }
            },
            [SortableField(id="id")],
            [{"id": {"order": SortOrder.asc}}],
        ),
        (
            {
                "mappings": {
                    "properties": {
                        "id": {"type": "keyword"},
                    }
                }
            },
            [SortableField(id="metadata.black", order=SortOrder.desc)],
            [{"metadata.black": {"order": SortOrder.desc}}],
        ),
    ],
)
def test_build_sort_configuration(index_schema, sort_cfg, expected_sort):
    builder = EsQueryBuilder()

    es_sort = builder.map_2_es_sort_configuration(sort=SortConfig(sort_by=sort_cfg), schema=index_schema)
    assert es_sort == expected_sort


def test_build_sort_with_wrong_field_name():
    builder = EsQueryBuilder()

    index_schema = {
        "mappings": {
            "properties": {
                "id": {"type": "keyword"},
            }
        }
    }

    with pytest.raises(Exception):
        builder.map_2_es_sort_configuration(schema=index_schema, sort=SortConfig(sort_by=[SortableField(id="wat?!")]))


def test_build_sort_without_sort_config():
    builder = EsQueryBuilder()
    index_schema = {
        "mappings": {
            "properties": {
                "id": {"type": "keyword"},
            }
        }
    }
    assert builder.map_2_es_sort_configuration(sort=SortConfig(), schema=index_schema) is None


def test_query_builder_with_query_range():
    es_query = EsQueryBuilder().map_2_es_query(
        schema=None,
        query=TextClassificationQuery(score=ScoreRange(range_from=10)),
    )
    assert es_query == {
        "query": {
            "bool": {
                "filter": {
                    "bool": {
                        "minimum_should_match": 1,
                        "should": [{"range": {"score": {"gte": 10.0}}}],
                    }
                },
                "must": {"match_all": {}},
            }
        }
    }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, Union


class Claims(dict):
    """Claims configuration for a single provider."""

    display_name: Union[str, Callable[[dict], Any]]
    identity: Union[str, Callable[[dict], Any]]
    picture: Union[str, Callable[[dict], Any]]
    email: Union[str, Callable[[dict], Any]]

    def __init__(self, seq=None, **kwargs) -> None:
        super().__init__(seq or {}, **kwargs)
        self["display_name"] = kwargs.get("display_name", self.get("display_name", "name"))
        self["identity"] = kwargs.get("identity", self.get("identity", "sub"))
        self["picture"] = kwargs.get("picture", self.get("picture", "picture"))
        self["email"] = kwargs.get("email", self.get("email", "email"))



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""fix suggestions type enum values

Revision ID: 1e629a913727
Revises: 3fc3c0839959
Create Date: 2023-07-24 12:47:11.715011

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1e629a913727"
down_revision = "3fc3c0839959"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()

    if bind.dialect.name == "postgresql":
        op.execute("ALTER TYPE suggestion_type_enum ADD VALUE IF NOT EXISTS 'model';")
        op.execute("ALTER TYPE suggestion_type_enum ADD VALUE IF NOT EXISTS 'human';")


def downgrade() -> None:
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import secrets
from typing import TYPE_CHECKING, List, Union
from uuid import UUID

from passlib.context import CryptContext
from sqlalchemy import exists, select
from sqlalchemy.orm import Session, selectinload

from argilla_server.enums import UserRole
from argilla_server.models import User, Workspace, WorkspaceUser
from argilla_server.schemas.v0.users import UserCreate
from argilla_server.schemas.v0.workspaces import WorkspaceCreate, WorkspaceUserCreate

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

_CRYPT_CONTEXT = CryptContext(schemes=["bcrypt"], deprecated="auto")


async def get_workspace_user_by_workspace_id_and_user_id(
    db: "AsyncSession", workspace_id: UUID, user_id: UUID
) -> Union[WorkspaceUser, None]:
    result = await db.execute(select(WorkspaceUser).filter_by(workspace_id=workspace_id, user_id=user_id))
    return result.scalar_one_or_none()


async def create_workspace_user(db: "AsyncSession", workspace_user_create: WorkspaceUserCreate) -> WorkspaceUser:
    workspace_user = await WorkspaceUser.create(
        db,
        workspace_id=workspace_user_create.workspace_id,
        user_id=workspace_user_create.user_id,
    )
    await db.refresh(workspace_user, attribute_names=["workspace", "user"])
    return workspace_user


async def delete_workspace_user(db: "AsyncSession", workspace_user: WorkspaceUser) -> WorkspaceUser:
    return await workspace_user.delete(db)


async def get_workspace_by_id(db: "AsyncSession", workspace_id: UUID) -> Workspace:
    return await Workspace.read(db, id=workspace_id)


async def get_workspace_by_name(db: "AsyncSession", workspace_name: str) -> Union[Workspace, None]:
    result = await db.execute(select(Workspace).filter_by(name=workspace_name))
    return result.scalar_one_or_none()


async def list_workspaces(db: "AsyncSession") -> List[Workspace]:
    result = await db.execute(select(Workspace).order_by(Workspace.inserted_at.asc()))
    return result.scalars().all()


async def list_workspaces_by_user_id(db: "AsyncSession", user_id: UUID) -> List[Workspace]:
    result = await db.execute(
        select(Workspace)
        .join(WorkspaceUser)
        .filter(WorkspaceUser.user_id == user_id)
        .order_by(Workspace.inserted_at.asc())
    )
    return result.scalars().all()


async def create_workspace(db: "AsyncSession", workspace_create: WorkspaceCreate) -> Workspace:
    return await Workspace.create(db, schema=workspace_create)


async def delete_workspace(db: "AsyncSession", workspace: Workspace):
    return await workspace.delete(db)


async def get_user_by_id(db: "AsyncSession", user_id: UUID) -> Union[User, None]:
    return await User.read(db, id=user_id)


async def user_exists(db: "AsyncSession", user_id: UUID) -> bool:
    return await db.scalar(select(exists().where(User.id == user_id)))


def get_user_by_username_sync(db: Session, username: str) -> Union[User, None]:
    return db.query(User).filter_by(username=username).first()


async def get_user_by_username(db: "AsyncSession", username: str) -> Union[User, None]:
    result = await db.execute(select(User).filter_by(username=username).options(selectinload(User.workspaces)))
    return result.scalar_one_or_none()


def get_user_by_api_key_sync(db: Session, api_key: str) -> Union[User, None]:
    return db.query(User).filter_by(api_key=api_key).first()


async def get_user_by_api_key(db: "AsyncSession", api_key: str) -> Union[User, None]:
    result = await db.execute(select(User).where(User.api_key == api_key).options(selectinload(User.workspaces)))
    return result.scalar_one_or_none()


async def list_users(db: "AsyncSession") -> List[User]:
    result = await db.execute(select(User).order_by(User.inserted_at.asc()).options(selectinload(User.workspaces)))
    return result.scalars().all()


async def create_user(db: "AsyncSession", user_create: UserCreate) -> User:
    async with db.begin_nested():
        user = await User.create(
            db,
            first_name=user_create.first_name,
            last_name=user_create.last_name,
            username=user_create.username,
            role=user_create.role,
            password_hash=hash_password(user_create.password),
            autocommit=False,
        )

        if user_create.workspaces:
            for workspace_name in user_create.workspaces:
                workspace = await get_workspace_by_name(db, workspace_name)
                if not workspace:
                    raise ValueError(f"Workspace '{workspace_name}' does not exist")
                await WorkspaceUser.create(
                    db,
                    workspace_id=workspace.id,
                    user_id=user.id,
                    autocommit=False,
                )

    await db.commit()

    return user


async def create_user_with_random_password(
    db,
    username: str,
    first_name: str,
    workspaces: List[str] = None,
    role: UserRole = UserRole.annotator,
) -> User:
    password = _generate_random_password()

    user_create = UserCreate(
        first_name=first_name, username=username, role=role, password=password, workspaces=workspaces
    )
    return await create_user(db, user_create)


async def delete_user(db: "AsyncSession", user: User) -> User:
    return await user.delete(db)


async def authenticate_user(db: "AsyncSession", username: str, password: str):
    user = await get_user_by_username(db, username)

    if user and verify_password(password, user.password_hash):
        return user
    elif user:
        return
    else:
        _CRYPT_CONTEXT.dummy_verify()


def hash_password(password: str) -> str:
    return _CRYPT_CONTEXT.hash(password)


def verify_password(password: str, password_hash: str) -> bool:
    return _CRYPT_CONTEXT.verify(password, password_hash)


def _generate_random_password() -> str:
    return secrets.token_urlsafe()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
from typing import List, Optional

import typer

from argilla_server.contexts import accounts
from argilla_server.database import AsyncSessionLocal
from argilla_server.models import User, UserRole
from argilla_server.pydantic_v1 import constr
from argilla_server.schemas.v0.users import USER_PASSWORD_MIN_LENGTH, UserCreate
from argilla_server.schemas.v0.workspaces import WorkspaceCreate

from .utils import get_or_new_workspace

USER_API_KEY_MIN_LENGTH = 8


class UserCreateForTask(UserCreate):
    api_key: Optional[constr(min_length=USER_API_KEY_MIN_LENGTH)]
    workspaces: Optional[List[WorkspaceCreate]]


def role_callback(value: str) -> str:
    try:
        return UserRole(value).value
    except ValueError:
        raise typer.BadParameter("Only Camila is allowed")


def password_callback(password: str = None) -> str:
    # if password is None:
    #     raise typer.BadParameter("Password must be specified.")
    # if len(password)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"Password must be at least {USER_PASSWORD_MIN_LENGTH} characters long.")
    return password


def api_key_callback(api_key: str) -> str:
    # if api_key and len(api_key)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"API key must a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters.")
    return api_key


async def _create(
    first_name: str,
    username: str,
    role: UserRole,
    password: str,
    last_name: Optional[str] = None,
    api_key: Optional[str] = None,
    workspace: List[str] = None,
):
    """Creates a new user in the Argilla database with provided parameters"""
    if workspace is None:
        workspace = []

    async with AsyncSessionLocal() as session:
        if await accounts.get_user_by_username(session, username):
            typer.echo(f"User with username {username!r} already exists in database. Skipping...")
            return

        if await accounts.get_user_by_api_key(session, api_key):
            typer.echo(f"User with api_key {api_key!r} already exists in database. Skipping...")
            return

        user_create = UserCreateForTask(
            first_name=first_name,
            last_name=last_name,
            username=username,
            role=role,
            password=password,
            api_key=api_key,
            workspaces=[WorkspaceCreate(name=workspace_name) for workspace_name in workspace],
        )

        user = await User.create(
            session,
            first_name=user_create.first_name,
            last_name=user_create.last_name,
            username=user_create.username,
            role=user_create.role,
            password_hash=accounts.hash_password(user_create.password),
            api_key=user_create.api_key,
            workspaces=[await get_or_new_workspace(session, workspace.name) for workspace in user_create.workspaces],
        )

        typer.echo("User successfully created:")
        typer.echo(f"• first_name: {user.first_name!r}")
        if user.last_name:
            typer.echo(f"• last_name: {user.last_name!r}")
        typer.echo(f"• username: {user.username!r}")
        typer.echo(f"• role: {user.role.value!r}")
        typer.echo(f"• api_key: {user.api_key!r}")
        typer.echo(f"• workspaces: {[workspace.name for workspace in user.workspaces]!r}")


def create(
    first_name: str = typer.Option(default=None, help="First name as a string."),
    username: str = typer.Option(
        default=None,
        prompt=True,
        help="Username as a lowercase string without spaces allowing letters, numbers, dashes and underscores.",
    ),
    role: UserRole = typer.Option(
        prompt=True,
        default=UserRole.annotator.value,
        show_default=True,
        help="Role for the user.",
    ),
    password: str = typer.Option(
        default=None,
        prompt=True,
        confirmation_prompt=True,
        hide_input=True,
        callback=password_callback,
        help=f"Password as a string with a minimum length of {USER_PASSWORD_MIN_LENGTH} characters.",
    ),
    last_name: str = typer.Option(default=None, help="Last name as a string."),
    api_key: Optional[str] = typer.Option(
        default=None,
        callback=api_key_callback,
        help=f"API key as a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters. If not specified a secure random API key will be generated",
    ),
    workspace: List[str] = typer.Option(
        default=[], help="A workspace that the user will be a member of (can be used multiple times)."
    ),
):
    asyncio.run(_create(first_name, username, role, password, last_name, api_key, workspace))


if __name__ == "__main__":
    typer.run(create)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings

import pytest
from argilla_server.daos.models.records import BaseRecordInDB
from argilla_server.settings import settings


def test_metadata_flatten():
    raw_data = {"a": 1, "b": "value"}

    record = BaseRecordInDB(metadata={"a": {"json": raw_data}})
    assert record.metadata == {"a.json.a": 1, "a.json.b": "value"}


def test_metadata_limit():
    long_value = "a" * (settings.metadata_field_length + 1)
    short_value = "a" * (settings.metadata_field_length - 1)

    with pytest.warns(expected_warning=UserWarning):
        BaseRecordInDB(metadata=dict(a=long_value))

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        BaseRecordInDB(metadata=dict(a=short_value))


def test_protected_metadata_without_truncation():
    long_value = "a" * (settings.metadata_field_length + 1)

    record = BaseRecordInDB(metadata=dict(a=long_value, _protected=long_value))
    assert record.metadata["a"] == long_value[: settings.metadata_field_length]
    assert record.metadata["_protected"] == long_value


def test_protected_metadata_with_json_dict():
    raw_data = {"a": 1, "b": "value"}

    record = BaseRecordInDB(metadata={"json": raw_data, "_json": raw_data})
    assert record.metadata == {"json.a": 1, "json.b": "value", "_json": {"a": 1, "b": "value"}}



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Optional, Set, Tuple

from argilla_server.commons.models import PredictionStatus, TaskType
from argilla_server.constants import DEFAULT_MAX_KEYWORD_LENGTH
from argilla_server.pydantic_v1 import BaseModel, Field, validator
from argilla_server.services.search.model import ServiceBaseRecordsQuery, ServiceScoreRange
from argilla_server.services.tasks.commons import ServiceBaseAnnotation, ServiceBaseRecord
from argilla_server.utils import SpanUtils

PREDICTED_MENTIONS_ES_FIELD_NAME = "predicted_mentions"
MENTIONS_ES_FIELD_NAME = "mentions"


class EntitySpan(BaseModel):
    """
    The tokens span for a labeled text.

    Entity spans will be defined between from start to end - 1

    Attributes:
    -----------

    start: int
        character start position
    end: int
        character end position, must be higher than the starting character.
    label: str
        the label related to tokens that conforms the entity span
    score:
        A higher score means, the model/annotator is more confident about its predicted/annotated entity.
    """

    start: int
    end: int
    label: str = Field(min_length=1, max_length=DEFAULT_MAX_KEYWORD_LENGTH)
    score: float = Field(default=1.0, ge=0.0, le=1.0)

    @validator("end")
    def check_span_offset(cls, end: int, values):
        """Validates span offset"""
        assert (
            end > values["start"]
        ), "End character cannot be placed before the starting character, it must be at least one character after."
        return end

    def __hash__(self):
        return hash(type(self)) + hash(self.__dict__.values())


class ServiceTokenClassificationAnnotation(ServiceBaseAnnotation):
    entities: List[EntitySpan] = Field(default_factory=list)
    score: Optional[float] = None


class ServiceTokenClassificationRecord(ServiceBaseRecord[ServiceTokenClassificationAnnotation]):
    tokens: List[str] = Field(min_items=1)
    text: str = Field()
    _raw_text: Optional[str] = Field(alias="raw_text")
    _span_utils: SpanUtils

    # TODO: review this.
    _predicted: Optional[PredictionStatus] = Field(alias="predicted")

    def __init__(self, **data):
        super().__init__(**data)

        self._span_utils = SpanUtils(self.text, self.tokens)

        if self.annotation:
            self._validate_spans(self.annotation)
        if self.prediction:
            self._validate_spans(self.prediction)

    def _validate_spans(self, annotation: ServiceTokenClassificationAnnotation):
        """Validates the spans with respect to the tokens.

        If necessary, also performs an automatic correction of the spans.

        Args:
            span_utils: Helper class to perform the checks.
            annotation: Contains the spans to validate.

        Raises:
            ValidationError: If spans are not valid or misaligned.
        """
        spans = [(ent.label, ent.start, ent.end) for ent in annotation.entities]
        try:
            self._span_utils.validate(spans)
        except ValueError:
            corrected_spans = self._span_utils.correct(spans)
            self._span_utils.validate(corrected_spans)
            for ent, span in zip(annotation.entities, corrected_spans):
                ent.start, ent.end = span[1], span[2]

    def task(cls) -> TaskType:
        """The record task type"""
        return TaskType.token_classification

    @property
    def span_utils(self) -> SpanUtils:
        """Utility class for span operations."""
        return self._span_utils

    @property
    def predicted(self) -> Optional[PredictionStatus]:
        if self.annotation and self.prediction:
            annotated_entities = self.annotation.entities
            predicted_entities = self.prediction.entities
            if len(annotated_entities) != len(predicted_entities):
                return PredictionStatus.KO

            for ann, pred in zip(annotated_entities, predicted_entities):
                if ann.start != pred.start or ann.end != pred.end or ann.label != pred.label:
                    return PredictionStatus.KO

            return PredictionStatus.OK
        return None

    @property
    def predicted_as(self) -> List[str]:
        return [ent.label for ent in self.predicted_entities()]

    @property
    def annotated_as(self) -> List[str]:
        return [ent.label for ent in self.annotated_entities()]

    @property
    def scores(self) -> List[float]:
        if not self.prediction:
            return []
        if self.prediction.score is not None:
            return [self.prediction.score]
        return [e.score for e in self.prediction.entities]

    def all_text(self) -> str:
        return self.text

    def predicted_mentions(self) -> List[Tuple[str, EntitySpan]]:
        return [
            (mention, entity) for mention, entity in self.__mentions_from_entities__(self.predicted_entities()).items()
        ]

    def annotated_mentions(self) -> List[Tuple[str, EntitySpan]]:
        return [
            (mention, entity) for mention, entity in self.__mentions_from_entities__(self.annotated_entities()).items()
        ]

    def annotated_entities(self) -> Set[EntitySpan]:
        """Shortcut for real annotated entities, if provided"""
        if self.annotation is None:
            return set()
        return set(self.annotation.entities)

    def predicted_entities(self) -> Set[EntitySpan]:
        """Predicted entities"""
        if self.prediction is None:
            return set()
        return set(self.prediction.entities)

    def __mentions_from_entities__(self, entities: Set[EntitySpan]) -> Dict[str, EntitySpan]:
        return {mention: entity for entity in entities for mention in [self.text[entity.start : entity.end]]}

    class Config:
        allow_population_by_field_name = True
        underscore_attrs_are_private = True


class ServiceTokenClassificationQuery(ServiceBaseRecordsQuery):
    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    score: Optional[ServiceScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Depends
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import accounts
from argilla_server.database import get_async_db
from argilla_server.errors import UnauthorizedError
from argilla_server.schemas.v0.authentication import Token, UserPasswordRequestForm
from argilla_server.security.authentication.jwt import JWT
from argilla_server.security.authentication.userinfo import UserInfo

router = APIRouter(tags=["Authentication"])


@router.post("/security/token", response_model=Token)
async def create_access_token(
    db: AsyncSession = Depends(get_async_db), form: UserPasswordRequestForm = Depends()
) -> Token:
    user = await accounts.authenticate_user(db, form.username, form.password)
    if not user:
        raise UnauthorizedError()

    token = JWT.create(UserInfo(username=user.username, name=user.first_name, role=user.role, identity=str(user.id)))

    return Token(access_token=token)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

import pytest
from httpx import AsyncClient

from tests.factories import DatasetFactory, RecordFactory, TextFieldFactory, VectorFactory, VectorSettingsFactory


@pytest.mark.asyncio
class TestSearchCurrentUserDatasetRecords:
    def url(self, dataset_id: UUID) -> str:
        return f"/api/v1/me/datasets/{dataset_id}/records/search"

    async def test_with_vector_query_using_record_without_vector(
        self, async_client: AsyncClient, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        await TextFieldFactory.create(name="input", dataset=dataset)
        vector_settings = await VectorSettingsFactory.create(name="vector", dimensions=3, dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)
        record_without_vector = await RecordFactory.create(dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings, record=record)

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "query": {
                    "vector": {
                        "name": vector_settings.name,
                        "record_id": str(record_without_vector.id),
                    },
                },
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Record `{record_without_vector.id}` does not have a vector for vector settings `{vector_settings.name}`"
        }

    async def test_with_invalid_filter(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "query": {
                    "text": {"q": "text", "field": "non-existent"},
                },
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "response", "question": "non-existent"},
                            "values": ["value-a"],
                        }
                    ],
                },
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        }

    async def test_with_invalid_sort(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "query": {
                    "text": {"q": "text", "field": "non-existent"},
                },
                "sort": [{"scope": {"entity": "response", "question": "non-existent"}, "order": "asc"}],
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        }



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

import pytest
from argilla_server.logging import ArgillaHandler, LoggingMixin


class LoggingForTest(LoggingMixin):
    """class for tests"""

    def __init__(self, a: int = 5):
        self.a = a

    def f(self):
        self.logger.warning("This is an warning %d", self.a)


class LoggingForTestChild(LoggingForTest):
    """class child"""

    pass


def test_logging_mixin_without_breaking_constructors():
    test = LoggingForTest()
    test.f()
    assert test.logger.name == f"{__name__}.{LoggingForTest.__name__}"

    child = LoggingForTestChild(a=10)
    child.f()
    assert child.logger.name == f"{__name__}.{LoggingForTestChild.__name__}"

    assert test.logger != child.logger

    another_child = LoggingForTestChild(a=15)
    # Check logger without call property method
    assert another_child.__getattribute__("__logger__") == child.logger


def test_logging_handler(mocker):
    mocker.patch.object(ArgillaHandler, "emit", autospec=True)
    handler = ArgillaHandler()

    logger = logging.getLogger(__name__)
    logger.handlers = [handler]
    logger.setLevel(logging.INFO)  # Force trace
    logger.info("This is a test")
    handler.emit.assert_called()


@pytest.mark.skip(reason="Failing temporally")
def test_configure_logging_call():
    # Ensure that the root logger uses the ArgillaHandler (RichHandler if rich is installed),
    # whereas the other loggers do not have handlers
    assert isinstance(logging.getLogger().handlers[0], ArgillaHandler)
    assert len(logging.getLogger("argilla").handlers) == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID, uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from httpx import AsyncClient

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    DatasetFactory,
    OwnerFactory,
    QuestionFactory,
    SuggestionFactory,
)


@pytest.mark.asyncio
class TestListDatasetRecordsSearchSuggestionsOptions:
    def url(self, dataset_id: UUID) -> str:
        return f"/api/v1/datasets/{dataset_id}/records/search/suggestions/options"

    async def test(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        question_a = await QuestionFactory.create(name="question-a", dataset=dataset)
        question_b = await QuestionFactory.create(name="question-b", dataset=dataset)
        question_c = await QuestionFactory.create(name="question-c", dataset=dataset)

        await SuggestionFactory.create(question=question_a)
        await SuggestionFactory.create(question=question_b, agent="agent-a")
        await SuggestionFactory.create(question=question_b, agent="agent-a")
        await SuggestionFactory.create(question=question_c, agent="agent-a")
        await SuggestionFactory.create(question=question_c, agent="agent-b")
        await SuggestionFactory.create(question=question_c, agent="agent-c")
        await SuggestionFactory.create(question=question_c, agent="agent-c")

        extra_dataset = await DatasetFactory.create()
        extra_question_a = await QuestionFactory.create(name="extra-question-a", dataset=extra_dataset)
        extra_question_b = await QuestionFactory.create(name="extra-question-b", dataset=extra_dataset)

        await SuggestionFactory.create(question=extra_question_a, agent="extra-agent-a")
        await SuggestionFactory.create(question=extra_question_b, agent="extra-agent-b")

        response = await async_client.get(self.url(dataset.id), headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "question": {
                        "id": str(question_a.id),
                        "name": "question-a",
                    },
                    "agents": [],
                },
                {
                    "question": {
                        "id": str(question_b.id),
                        "name": "question-b",
                    },
                    "agents": ["agent-a"],
                },
                {
                    "question": {
                        "id": str(question_c.id),
                        "name": "question-c",
                    },
                    "agents": ["agent-a", "agent-b", "agent-c"],
                },
            ]
        }

    async def test_with_dataset_without_questions(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.get(self.url(dataset.id), headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {"items": []}

    async def test_with_dataset_without_suggestions(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        question_a = await QuestionFactory.create(name="question-a", dataset=dataset)
        question_b = await QuestionFactory.create(name="question-b", dataset=dataset)
        question_c = await QuestionFactory.create(name="question-c", dataset=dataset)

        extra_dataset = await DatasetFactory.create()
        extra_question_a = await QuestionFactory.create(name="extra-question-a", dataset=extra_dataset)
        extra_question_b = await QuestionFactory.create(name="extra-question-b", dataset=extra_dataset)

        await SuggestionFactory.create(question=extra_question_a, agent="extra-agent-a")
        await SuggestionFactory.create(question=extra_question_b, agent="extra-agent-b")

        response = await async_client.get(self.url(dataset.id), headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "question": {
                        "id": str(question_a.id),
                        "name": question_a.name,
                    },
                    "agents": [],
                },
                {
                    "question": {
                        "id": str(question_b.id),
                        "name": question_b.name,
                    },
                    "agents": [],
                },
                {
                    "question": {
                        "id": str(question_c.id),
                        "name": question_c.name,
                    },
                    "agents": [],
                },
            ]
        }

    async def test_as_owner(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        owner = await OwnerFactory.create(workspaces=[dataset.workspace])

        response = await async_client.get(self.url(dataset.id), headers={API_KEY_HEADER_NAME: owner.api_key})

        assert response.status_code == 200

    async def test_as_admin(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        admin = await AdminFactory.create(workspaces=[dataset.workspace])

        response = await async_client.get(self.url(dataset.id), headers={API_KEY_HEADER_NAME: admin.api_key})

        assert response.status_code == 200

    async def test_as_annotator(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        annotator = await AnnotatorFactory.create(workspaces=[dataset.workspace])

        response = await async_client.get(self.url(dataset.id), headers={API_KEY_HEADER_NAME: annotator.api_key})

        assert response.status_code == 200

    async def test_as_user_from_different_workspace(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        annotator = await AnnotatorFactory.create()

        response = await async_client.get(self.url(dataset.id), headers={API_KEY_HEADER_NAME: annotator.api_key})

        assert response.status_code == 403

    async def test_with_non_existent_dataset(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.get(self.url(uuid4()), headers=owner_auth_header)

        assert response.status_code == 404



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""create vectors table

Revision ID: bda6fe24314e
Revises: 7850ab5b42d9
Create Date: 2023-09-21 15:41:59.642012

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bda6fe24314e"
down_revision = "7850ab5b42d9"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "vectors",
        sa.Column("value", sa.JSON(), nullable=False),
        sa.Column("record_id", sa.Uuid(), nullable=False),
        sa.Column("vector_settings_id", sa.Uuid(), nullable=False),
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("inserted_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["record_id"], ["records.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["vector_settings_id"], ["vectors_settings.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("record_id", "vector_settings_id", name="vector_record_id_vector_settings_id_uq"),
    )
    op.create_index(op.f("ix_vectors_record_id"), "vectors", ["record_id"], unique=False)
    op.create_index(op.f("ix_vectors_vector_settings_id"), "vectors", ["vector_settings_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_vectors_vector_settings_id"), table_name="vectors")
    op.drop_index(op.f("ix_vectors_record_id"), table_name="vectors")
    op.drop_table("vectors")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import accounts, datasets
from argilla_server.database import get_async_db
from argilla_server.models import User
from argilla_server.policies import WorkspacePolicyV1, authorize
from argilla_server.schemas.v1.workspaces import Workspace, Workspaces
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService

router = APIRouter(tags=["workspaces"])


@router.get("/workspaces/{workspace_id}", response_model=Workspace)
async def get_workspace(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, WorkspacePolicyV1.get(workspace_id))

    workspace = await accounts.get_workspace_by_id(db, workspace_id)
    if not workspace:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Workspace with id `{workspace_id}` not found",
        )

    return workspace


@router.delete("/workspaces/{workspace_id}", response_model=Workspace)
async def delete_workspace(
    *,
    db: AsyncSession = Depends(get_async_db),
    datasets_service: DatasetsService = Depends(DatasetsService.get_instance),
    workspace_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, WorkspacePolicyV1.delete)

    workspace = await accounts.get_workspace_by_id(db, workspace_id)
    if not workspace:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Workspace with id `{workspace_id}` not found",
        )

    if await datasets.list_datasets_by_workspace_id(db, workspace_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Cannot delete the workspace {workspace_id}. This workspace has some feedback datasets linked",
        )

    if await datasets_service.list(current_user, workspaces=[workspace.name]):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Cannot delete the workspace {workspace_id}. This workspace has some datasets linked",
        )

    return await accounts.delete_workspace(db, workspace)


@router.get("/me/workspaces", response_model=Workspaces)
async def list_workspaces_me(
    *, db: AsyncSession = Depends(get_async_db), current_user: User = Security(auth.get_current_user)
) -> Workspaces:
    await authorize(current_user, WorkspacePolicyV1.list_workspaces_me)

    if current_user.is_owner:
        workspaces = await accounts.list_workspaces(db)
    else:
        workspaces = await accounts.list_workspaces_by_user_id(db, current_user.id)

    return Workspaces(items=workspaces)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import pytest
from argilla_server.contexts import accounts
from argilla_server.models import User, UserRole, Workspace
from click.testing import CliRunner
from typer import Typer

from tests.factories import UserSyncFactory, WorkspaceSyncFactory

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@pytest.mark.parametrize("role_string", ["owner", "admin", "annotator"])
def test_create(sync_db: "Session", cli_runner: CliRunner, cli: Typer, role_string: str):
    result = cli_runner.invoke(
        cli,
        f"database users create --first-name first-name --username username --role {role_string} "
        "--password 12345678 --workspace workspace",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 1

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert user.first_name == "first-name"
    assert user.username == "username"
    assert user.role.value == UserRole(role_string).value
    assert accounts.verify_password("12345678", user.password_hash)
    assert user.api_key
    assert [ws.name for ws in user.workspaces] == ["workspace"]


def test_create_with_default_role(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --password 12345678",
        input="\n",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert user.role == UserRole.annotator


@pytest.mark.parametrize("role_string", ["owner", "admin", "annotator"])
def test_create_with_input_role(sync_db: "Session", cli_runner: CliRunner, cli: Typer, role_string: str):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --password 12345678",
        input=f"{role_string}\n",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert user.role.value == UserRole(role_string).value


def test_create_with_invalid_role(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role bad_role "
        "--password 12345678 --workspace workspace",
    )

    assert result.exit_code == 2
    assert sync_db.query(User).count() == 0
    assert sync_db.query(Workspace).count() == 0


def test_create_with_input_password(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner",
        input="12345678\n12345678\n",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert accounts.verify_password("12345678", user.password_hash)


def test_create_with_invalid_password(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli, "database users create --first-name first-name --username username --password 1234 --role owner"
    )

    assert result.exit_code == 1
    assert sync_db.query(User).count() == 0
    assert sync_db.query(Workspace).count() == 0


def test_create_with_input_username(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli, "database users create --first-name first-name --password 12345678", input="username\n"
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    assert sync_db.query(User).filter_by(username="username").first()


def test_create_with_invalid_username(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username -Invalid-Username --password 12345678 --role owner",
    )

    assert result.exit_code == 1
    assert sync_db.query(User).count() == 0
    assert sync_db.query(Workspace).count() == 0


def test_create_with_existing_username(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    UserSyncFactory.create(username="username")

    result = cli_runner.invoke(
        cli, "database users create --first-name first-name --username username --role owner --password 12345678"
    )

    assert result.exit_code == 0
    assert "username" in result.output
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0


def test_create_with_last_name(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --last-name last-name --username username --password 12345678 --role owner",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert user.last_name == "last-name"


def test_create_with_api_key(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 --api-key abcdefgh",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert user.api_key == "abcdefgh"


def test_create_with_invalid_api_key(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 --api-key abc",
    )

    assert result.exit_code == 1
    assert sync_db.query(User).count() == 0
    assert sync_db.query(Workspace).count() == 0


def test_create_with_existing_api_key(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    UserSyncFactory.create(api_key="abcdefgh")

    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 --api-key abcdefgh",
    )

    assert result.exit_code == 0
    assert "abcdefgh" in result.output
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 0


def test_create_with_multiple_workspaces(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 "
        "--workspace workspace-a --workspace workspace-b",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 2

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert [ws.name for ws in user.workspaces] == ["workspace-a", "workspace-b"]


def test_create_with_existent_workspaces(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    WorkspaceSyncFactory.create(name="workspace-a")
    WorkspaceSyncFactory.create(name="workspace-b")

    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 "
        "--workspace workspace-a --workspace workspace-b --workspace workspace-c",
    )

    assert result.exit_code == 0
    assert sync_db.query(User).count() == 1
    assert sync_db.query(Workspace).count() == 3

    user = sync_db.query(User).filter_by(username="username").first()
    assert user
    assert [ws.name for ws in user.workspaces] == ["workspace-a", "workspace-b", "workspace-c"]


def test_create_with_invalid_workspaces(sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    result = cli_runner.invoke(
        cli,
        "database users create --first-name first-name --username username --role owner --password 12345678 "
        "--workspace workspace-a --workspace 'invalid workspace' --workspace workspace-c",
    )

    assert result.exit_code == 1
    assert sync_db.query(User).count() == 0
    assert sync_db.query(Workspace).count() == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""add last_activity_at to datasets table

Revision ID: 84f6b9ff6076
Revises: b8458008b60e
Create Date: 2023-10-19 16:06:16.097130

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "84f6b9ff6076"
down_revision = "b8458008b60e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("datasets", sa.Column("last_activity_at", sa.DateTime(), nullable=True))
    op.execute("UPDATE datasets SET last_activity_at = updated_at")
    with op.batch_alter_table("datasets") as batch_op:
        batch_op.alter_column("last_activity_at", nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("datasets", "last_activity_at")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.models.datasets import BaseDatasetDB
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.errors import MissingDatasetRecordsError


def test_raise_proper_error():
    dao = DatasetRecordsDAO.get_instance(GenericElasticEngineBackend.get_instance())
    with pytest.raises(MissingDatasetRecordsError):
        dao.search_records(
            dataset=BaseDatasetDB(
                name="mock-notfound",
                workspace="workspace-name",
                task=TaskType.text_classification,
            )
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.security.authentication.oauth2 import OAuth2Settings


class TestOAuth2Settings:
    def test_configure_unsupported_provider(self):
        with pytest.raises(ValueError):
            OAuth2Settings.from_dict({"providers": [{"name": "unsupported"}]})

    def test_configure_github_provider(self):
        settings = OAuth2Settings.from_dict(
            {
                "providers": [
                    {
                        "name": "github",
                        "client_id": "github_client_id",
                        "client_secret": "github_client_secret",
                        "scope": "user:email",
                    }
                ]
            }
        )
        github_provider = settings.providers["github"]

        assert github_provider.name == "github"
        assert github_provider.client_id == "github_client_id"
        assert github_provider.client_secret == "github_client_secret"
        assert github_provider.scope == ["user:email"]

    def test_configure_huggingface_provider(self):
        settings = OAuth2Settings.from_dict(
            {
                "providers": [
                    {
                        "name": "huggingface",
                        "client_id": "huggingface_client_id",
                        "client_secret": "huggingface_client_secret",
                        "scope": "openid profile email",
                    }
                ]
            }
        )
        huggingface_provider = settings.providers["huggingface"]

        assert huggingface_provider.name == "huggingface"
        assert huggingface_provider.client_id == "huggingface_client_id"
        assert huggingface_provider.client_secret == "huggingface_client_secret"
        assert huggingface_provider.scope == ["openid", "profile", "email"]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from typing_extensions import Annotated

from argilla_server.models import QuestionType
from argilla_server.pydantic_v1 import BaseModel, Field, PositiveInt, conlist, constr, root_validator, validator
from argilla_server.schemas.base import UpdateSchema

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


QUESTION_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
QUESTION_CREATE_NAME_MIN_LENGTH = 1
QUESTION_CREATE_NAME_MAX_LENGTH = 200

QUESTION_CREATE_TITLE_MIN_LENGTH = 1
QUESTION_CREATE_TITLE_MAX_LENGTH = 500

QUESTION_CREATE_DESCRIPTION_MIN_LENGTH = 1
QUESTION_CREATE_DESCRIPTION_MAX_LENGTH = 1000

VALUE_TEXT_OPTION_VALUE_MIN_LENGTH = 1
VALUE_TEXT_OPTION_VALUE_MAX_LENGTH = 200
VALUE_TEXT_OPTION_TEXT_MIN_LENGTH = 1
VALUE_TEXT_OPTION_TEXT_MAX_LENGTH = 500
VALUE_TEXT_OPTION_DESCRIPTION_MIN_LENGTH = 1
VALUE_TEXT_OPTION_DESCRIPTION_MAX_LENGTH = 1000

LABEL_SELECTION_OPTIONS_MIN_ITEMS = 2
LABEL_SELECTION_OPTIONS_MAX_ITEMS = 250
LABEL_SELECTION_MIN_VISIBLE_OPTIONS = 3

RANKING_OPTIONS_MIN_ITEMS = 2
RANKING_OPTIONS_MAX_ITEMS = 50

RATING_OPTIONS_MIN_ITEMS = 2
RATING_OPTIONS_MAX_ITEMS = 10
RATING_LOWER_VALUE_ALLOWED = 1
RATING_UPPER_VALUE_ALLOWED = 10


class TextQuestionSettings(BaseModel):
    type: Literal[QuestionType.text]
    use_markdown: bool = False


class RatingQuestionSettingsOption(BaseModel):
    value: int


class RatingQuestionSettings(BaseModel):
    type: Literal[QuestionType.rating]
    options: conlist(item_type=RatingQuestionSettingsOption)


class ValueTextQuestionSettingsOption(BaseModel):
    value: str
    text: str
    description: Optional[str] = None


class LabelSelectionQuestionSettings(BaseModel):
    type: Literal[QuestionType.label_selection]
    options: conlist(item_type=ValueTextQuestionSettingsOption)
    visible_options: Optional[int] = None


class MultiLabelSelectionQuestionSettings(LabelSelectionQuestionSettings):
    type: Literal[QuestionType.multi_label_selection]


class RankingQuestionSettings(BaseModel):
    type: Literal[QuestionType.ranking]
    options: conlist(item_type=ValueTextQuestionSettingsOption)


QuestionSettings = Annotated[
    Union[
        TextQuestionSettings,
        RatingQuestionSettings,
        LabelSelectionQuestionSettings,
        MultiLabelSelectionQuestionSettings,
        RankingQuestionSettings,
    ],
    Field(..., discriminator="type"),
]


class TextQuestionSettingsUpdate(UpdateSchema):
    type: Literal[QuestionType.text]
    use_markdown: Optional[bool]

    __non_nullable_fields__ = {"use_markdown"}


class RatingQuestionSettingsUpdate(UpdateSchema):
    type: Literal[QuestionType.rating]


class LabelSelectionSettingsUpdate(UpdateSchema):
    type: Literal[QuestionType.label_selection]
    visible_options: Optional[PositiveInt]


class MultiLabelSelectionQuestionSettingsUpdate(LabelSelectionSettingsUpdate):
    type: Literal[QuestionType.multi_label_selection]


class RankingQuestionSettingsUpdate(UpdateSchema):
    type: Literal[QuestionType.ranking]


QuestionSettingsUpdate = Annotated[
    Union[
        TextQuestionSettingsUpdate,
        RatingQuestionSettingsUpdate,
        LabelSelectionSettingsUpdate,
        MultiLabelSelectionQuestionSettingsUpdate,
        RankingQuestionSettingsUpdate,
    ],
    Field(..., discriminator="type"),
]


QuestionName = Annotated[
    constr(
        regex=QUESTION_CREATE_NAME_REGEX,
        min_length=QUESTION_CREATE_NAME_MIN_LENGTH,
        max_length=QUESTION_CREATE_NAME_MAX_LENGTH,
    ),
    Field(..., description="The name of the question"),
]


QuestionTitle = Annotated[
    constr(
        min_length=QUESTION_CREATE_TITLE_MIN_LENGTH,
        max_length=QUESTION_CREATE_TITLE_MAX_LENGTH,
    ),
    Field(..., description="The title of the question"),
]


QuestionDescription = Annotated[
    constr(
        min_length=QUESTION_CREATE_DESCRIPTION_MIN_LENGTH,
        max_length=QUESTION_CREATE_DESCRIPTION_MAX_LENGTH,
    ),
    Field(..., description="The description of the question"),
]


class UniqueValuesCheckerMixin(BaseModel):
    @root_validator(skip_on_failure=True)
    def check_unique_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        options = values.get("options", [])
        seen = set()
        duplicates = set()
        for option in options:
            if option.value in seen:
                duplicates.add(option.value)
            else:
                seen.add(option.value)
        if duplicates:
            raise ValueError(f"Option values must be unique, found duplicates: {duplicates}")
        return values


class TextQuestionSettingsCreate(BaseModel):
    type: Literal[QuestionType.text]
    use_markdown: bool = False


class RatingQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.rating]
    options: conlist(
        item_type=RatingQuestionSettingsOption,
        min_items=RATING_OPTIONS_MIN_ITEMS,
        max_items=RATING_OPTIONS_MAX_ITEMS,
    )

    @validator("options")
    def check_option_value_range(cls, options: List[RatingQuestionSettingsOption]):
        """Validator to control all values are in allowed range 1 <= x <= 10"""
        for option in options:
            if not RATING_LOWER_VALUE_ALLOWED <= option.value <= RATING_UPPER_VALUE_ALLOWED:
                raise ValueError(
                    f"Option value {option.value!r} out of range "
                    f"[{RATING_LOWER_VALUE_ALLOWED!r}, {RATING_UPPER_VALUE_ALLOWED!r}]"
                )
        return options


class ValueTextQuestionSettingsOptionCreate(BaseModel):
    value: constr(
        min_length=VALUE_TEXT_OPTION_VALUE_MIN_LENGTH,
        max_length=VALUE_TEXT_OPTION_VALUE_MAX_LENGTH,
    )
    text: constr(
        min_length=VALUE_TEXT_OPTION_TEXT_MIN_LENGTH,
        max_length=VALUE_TEXT_OPTION_TEXT_MAX_LENGTH,
    )
    description: Optional[
        constr(
            min_length=VALUE_TEXT_OPTION_DESCRIPTION_MIN_LENGTH,
            max_length=VALUE_TEXT_OPTION_DESCRIPTION_MAX_LENGTH,
        )
    ] = None


class LabelSelectionQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.label_selection]
    options: conlist(
        item_type=ValueTextQuestionSettingsOptionCreate,
        min_items=LABEL_SELECTION_OPTIONS_MIN_ITEMS,
        max_items=LABEL_SELECTION_OPTIONS_MAX_ITEMS,
    )
    visible_options: Optional[int] = Field(None, ge=LABEL_SELECTION_MIN_VISIBLE_OPTIONS)

    @root_validator(skip_on_failure=True)
    def check_visible_options_value(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        visible_options = values.get("visible_options")
        if visible_options is not None:
            num_options = len(values["options"])
            if visible_options > num_options:
                raise ValueError(
                    "The value for 'visible_options' must be less or equal to the number of items in 'options'"
                    f" ({num_options})"
                )
        return values


class MultiLabelSelectionQuestionSettingsCreate(LabelSelectionQuestionSettingsCreate):
    type: Literal[QuestionType.multi_label_selection]


class RankingQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.ranking]
    options: conlist(
        item_type=ValueTextQuestionSettingsOptionCreate,
        min_items=RANKING_OPTIONS_MIN_ITEMS,
        max_items=RANKING_OPTIONS_MAX_ITEMS,
    )


QuestionSettingsCreate = Annotated[
    Union[
        TextQuestionSettingsCreate,
        RatingQuestionSettingsCreate,
        LabelSelectionQuestionSettingsCreate,
        MultiLabelSelectionQuestionSettingsCreate,
        RankingQuestionSettingsCreate,
    ],
    Field(discriminator="type"),
]


class Question(BaseModel):
    id: UUID
    name: str
    title: str
    description: Optional[str]
    required: bool
    settings: QuestionSettings
    dataset_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Questions(BaseModel):
    items: List[Question]


class QuestionCreate(BaseModel):
    name: QuestionName
    title: QuestionTitle
    description: Optional[QuestionDescription]
    required: Optional[bool]
    settings: QuestionSettingsCreate


class QuestionUpdate(UpdateSchema):
    title: Optional[QuestionTitle]
    description: Optional[QuestionDescription]
    settings: Optional[QuestionSettingsUpdate]

    __non_nullable_fields__ = {"title", "settings"}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""create metadata_properties table

Revision ID: 7cbcccf8b57a
Revises: 1e629a913727
Create Date: 2023-09-22 11:40:07.700301

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7cbcccf8b57a"
down_revision = "1e629a913727"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "metadata_properties",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("title", sa.Text(), nullable=False),
        sa.Column("settings", sa.JSON(), nullable=False),
        sa.Column("allowed_roles", sa.JSON(), server_default="[]", nullable=False),
        sa.Column("dataset_id", sa.Uuid(), nullable=False),
        sa.Column("inserted_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["datasets.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name", "dataset_id", name="metadata_property_name_dataset_id_uq"),
    )
    op.create_index(op.f("ix_metadata_properties_dataset_id"), "metadata_properties", ["dataset_id"], unique=False)
    op.create_index(op.f("ix_metadata_properties_name"), "metadata_properties", ["name"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_metadata_properties_name"), table_name="metadata_properties")
    op.drop_index(op.f("ix_metadata_properties_dataset_id"), table_name="metadata_properties")
    op.drop_table("metadata_properties")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.fixture(autouse=True)
def mock_session_local(mocker: "MockerFixture", async_db_proxy: "AsyncSession") -> None:
    mocker.patch("argilla_server.cli.database.users.create.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.update.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.create_default.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.migrate.AsyncSessionLocal", return_value=async_db_proxy)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture
    from typer import Typer


def test_start_command(cli_runner: "CliRunner", cli: "Typer", mocker: "MockerFixture") -> None:
    uvicorn_run_mock = mocker.patch("uvicorn.run")
    result = cli_runner.invoke(cli, "start --host 1.1.1.1 --port 6899 --no-access-log")

    assert result.exit_code == 0
    uvicorn_run_mock.assert_called_once_with("argilla_server:app", host="1.1.1.1", port=6899, access_log=False)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass

from fastapi import Header, Path, Query

from argilla_server.constants import ES_INDEX_REGEX_PATTERN, WORKSPACE_HEADER_NAME
from argilla_server.errors import BadRequestError, MissingInputParamError

DATASET_NAME_PATH_PARAM = Path(..., regex=ES_INDEX_REGEX_PATTERN, description="The dataset name")


@dataclass
class RequestPagination:
    """Query pagination params"""

    limit: int = Query(50, gte=0, le=1000, description="Response records limit")
    from_: int = Query(0, ge=0, le=10000, alias="from", description="Record sequence from")


@dataclass
class OptionalWorkspaceRequestDependency:
    """Common task query dependencies"""

    _description = (
        "The workspace where dataset belongs to. A valid workspace name should be provided. "
        "If not provided, the request will raise a 400 response error."
    )

    __workspace_header__: str = Header(None, alias=WORKSPACE_HEADER_NAME, deprecated=True, description=_description)

    __workspace_param__: str = Query(None, alias="workspace", description=_description)

    @property
    def workspace(self) -> str:
        """Return read workspace. Query param prior to header param"""
        return self.__workspace_param__ or self.__workspace_header__


@dataclass
class CommonTaskHandlerDependencies(OptionalWorkspaceRequestDependency):
    @property
    def workspace(self) -> str:
        workspace = super().workspace
        if not workspace:
            raise MissingInputParamError("A workspace must be provided")
        return workspace



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Optional, Union

from argilla_server.daos.models.datasets import EmbeddingsConfig
from argilla_server.pydantic_v1 import BaseModel, Field, validator


class AbstractDatasetSettings(BaseModel):
    vectors: Optional[Dict[str, Union[int, EmbeddingsConfig]]] = Field(
        default=None,
        description="The vectors configuration",
    )


class LabelsSchema(BaseModel):
    class Schema(BaseModel):
        id: str = Field(description="The label id")
        name: str = Field(description="The label name")
        description: Optional[str] = Field(None, description="The label description")

    labels: Union[List[str], List[Schema]] = Field(description="A set of labels")

    @validator("labels")
    def normalize_labels(cls, labels):
        """
        Labels schema accept a list of strings. Those string will be converted
        into ``LabelsSchema.Schema`` objects
        """
        if not labels:
            return labels
        if isinstance(labels[0], str):
            return [LabelsSchema.Schema(id=label, name=label) for label in labels]
        return labels


class WithLabelsSchemaSettings(AbstractDatasetSettings):
    label_schema: Optional[LabelsSchema] = Field(
        None,
        description="The dataset labels schema",
    )


class TextClassificationSettings(WithLabelsSchemaSettings):
    pass


class TokenClassificationSettings(WithLabelsSchemaSettings):
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.search_engine import ElasticSearchEngine
from argilla_server.search_engine.commons import ALL_RESPONSES_STATUSES_FIELD, es_index_name_for_dataset
from argilla_server.settings import settings
from opensearchpy import OpenSearch
from sqlalchemy.ext.asyncio import AsyncSession

from tests.factories import DatasetFactory, VectorSettingsFactory
from tests.unit.search_engine.test_commons import refresh_dataset


@pytest.mark.asyncio
@pytest.mark.skipif(not settings.search_engine == "elasticsearch", reason="Running on elasticsearch engine")
class TestElasticSearchEngine:
    async def test_create_index_for_dataset(
        self, elasticsearch_engine: ElasticSearchEngine, db: AsyncSession, opensearch: OpenSearch
    ):
        dataset = await DatasetFactory.create()

        await refresh_dataset(dataset)
        await elasticsearch_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"] == {
            "dynamic": "strict",
            "dynamic_templates": [
                {
                    "status_responses": {
                        "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                        "path_match": "responses.*.status",
                    }
                }
            ],
            "properties": {
                "id": {"type": "keyword"},
                "inserted_at": {"type": "date_nanos"},
                "updated_at": {"type": "date_nanos"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},
                "responses": {"dynamic": "true", "type": "object"},
                "metadata": {"dynamic": "false", "type": "object"},
            },
        }

        assert index["settings"]["index"]["number_of_shards"] == str(elasticsearch_engine.number_of_shards)
        assert index["settings"]["index"]["number_of_replicas"] == str(elasticsearch_engine.number_of_replicas)

    async def test_create_dataset_index_with_vectors(self, search_engine: ElasticSearchEngine, opensearch: OpenSearch):
        vectors_settings = await VectorSettingsFactory.create_batch(5)
        dataset = await DatasetFactory.create(vectors_settings=vectors_settings)

        await refresh_dataset(dataset)
        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"]["properties"]["vectors"]["properties"] == {
            str(settings.id): {
                "type": "dense_vector",
                "dims": settings.dimensions,
                "index": True,
                "similarity": "cosine",
            }
            for settings in vectors_settings
        }

    async def test_create_index_with_existing_index(self, search_engine: ElasticSearchEngine, opensearch: OpenSearch):
        from elasticsearch8 import RequestError

        dataset = await DatasetFactory.create()

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        with pytest.raises(RequestError, match="resource_already_exists_exception"):
            await search_engine.create_index(dataset)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

import pytest
from argilla_server.apis.v1.handlers.datasets.datasets import CREATE_DATASET_VECTOR_SETTINGS_MAX_COUNT
from httpx import AsyncClient

from tests.factories import DatasetFactory, VectorSettingsFactory


@pytest.mark.asyncio
class TestCreateDatasetVectorSettings:
    def url(self, dataset_id: UUID) -> str:
        return f"/api/v1/datasets/{dataset_id}/vectors-settings"

    async def test_with_maximum_number_of_vector_settings_reached(
        self, async_client: AsyncClient, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        await VectorSettingsFactory.create_batch(CREATE_DATASET_VECTOR_SETTINGS_MAX_COUNT, dataset=dataset)

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "name": "name",
                "title": "title",
                "dimensions": 3,
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"The maximum number of vector settings has been reached for dataset with id `{dataset.id}`"
        }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import UserRole
from argilla_server.schemas.v1.vector_settings import VECTOR_SETTINGS_CREATE_TITLE_MAX_LENGTH

from tests.factories import AdminFactory, AnnotatorFactory, UserFactory, VectorSettingsFactory

if TYPE_CHECKING:
    from httpx import AsyncClient


@pytest.mark.asyncio
class TestSuiteVectorsSettings:
    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_update_vector_settings(self, async_client: "AsyncClient", role: UserRole):
        vector_settings = await VectorSettingsFactory.create()
        user = await UserFactory.create(role=role, workspaces=[vector_settings.dataset.workspace])

        response = await async_client.patch(
            f"/api/v1/vectors-settings/{vector_settings.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={"title": "New Title"},
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(vector_settings.id),
            "name": vector_settings.name,
            "title": "New Title",
            "dimensions": vector_settings.dimensions,
            "dataset_id": str(vector_settings.dataset_id),
            "inserted_at": vector_settings.inserted_at.isoformat(),
            "updated_at": vector_settings.updated_at.isoformat(),
        }

        assert vector_settings.title == "New Title"

    @pytest.mark.parametrize("title", [None, "", "t" * (VECTOR_SETTINGS_CREATE_TITLE_MAX_LENGTH + 1)])
    async def test_update_vector_settings_with_invalid_title(
        self, async_client: "AsyncClient", owner_auth_header: dict, title: str
    ):
        vector_settings = await VectorSettingsFactory.create()

        response = await async_client.patch(
            f"/api/v1/vectors-settings/{vector_settings.id}",
            headers=owner_auth_header,
            json={"title": title},
        )

        assert response.status_code == 422

    async def test_update_vector_settings_with_title_as_none(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        vector_settings = await VectorSettingsFactory.create()

        response = await async_client.patch(
            f"/api/v1/vectors-settings/{vector_settings.id}",
            headers=owner_auth_header,
            json={"title": None},
        )

        assert response.status_code == 422

    async def test_update_vector_settings_non_existent(self, async_client: "AsyncClient", owner_auth_header: dict):
        response = await async_client.patch(
            f"/api/v1/vectors-settings/{uuid4()}",
            headers=owner_auth_header,
            json={"title": "New Title"},
        )

        assert response.status_code == 404

    async def test_update_vector_settings_as_admin_from_different_workspace(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        vector_settings = await VectorSettingsFactory.create()
        user = await AdminFactory.create()

        response = await async_client.patch(
            f"/api/v1/vectors-settings/{vector_settings.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={"title": "New Title"},
        )

        assert response.status_code == 403

    async def test_update_vector_settings_as_annotator(self, async_client: "AsyncClient", owner_auth_header: dict):
        vector_settings = await VectorSettingsFactory.create()
        user = await UserFactory.create()

        response = await async_client.patch(
            f"/api/v1/vectors-settings/{vector_settings.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={"title": "New Title"},
        )

        assert response.status_code == 403

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_delete_vector_settings(self, async_client: "AsyncClient", role: UserRole):
        vector_settings = await VectorSettingsFactory.create()
        user = await UserFactory.create(role=role, workspaces=[vector_settings.dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/vectors-settings/{vector_settings.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(vector_settings.id),
            "name": vector_settings.name,
            "title": vector_settings.title,
            "dimensions": vector_settings.dimensions,
            "dataset_id": str(vector_settings.dataset_id),
            "inserted_at": vector_settings.inserted_at.isoformat(),
            "updated_at": vector_settings.updated_at.isoformat(),
        }

    async def test_delete_vector_settings_non_existing(self, async_client: "AsyncClient", owner_auth_header: dict):
        response = await async_client.delete(f"/api/v1/vectors-settings/{uuid4()}", headers=owner_auth_header)

        assert response.status_code == 404

    async def test_delete_vector_settings_as_annotator(self, async_client: "AsyncClient"):
        vector_settings = await VectorSettingsFactory.create()
        annotator = await AnnotatorFactory.create(workspaces=[vector_settings.dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/vectors-settings/{vector_settings.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )

        assert response.status_code == 403

    async def test_delete_vector_settings_as_admin_from_another_workspace(self, async_client: "AsyncClient"):
        vector_settings = await VectorSettingsFactory.create()
        admin = await AdminFactory.create()

        response = await async_client.delete(
            f"/api/v1/vectors-settings/{vector_settings.id}", headers={API_KEY_HEADER_NAME: admin.api_key}
        )

        assert response.status_code == 403

    async def test_delete_vector_settings_without_authentication(self, async_client: "AsyncClient"):
        vector_settings = await VectorSettingsFactory.create()

        response = await async_client.delete(f"/api/v1/vectors-settings/{vector_settings.id}")

        assert response.status_code == 401



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import contextlib
import glob
import inspect
import logging
import os
import shutil
import tempfile
from pathlib import Path

import backoff
from brotli_asgi import BrotliMiddleware
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import RedirectResponse

from argilla_server import helpers
from argilla_server._version import __version__ as argilla_version
from argilla_server.constants import DEFAULT_API_KEY, DEFAULT_PASSWORD, DEFAULT_USERNAME
from argilla_server.contexts import accounts
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.backend.base import GenericSearchError
from argilla_server.daos.datasets import DatasetsDAO
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.database import get_async_db
from argilla_server.errors import (
    APIErrorHandler,
    BadRequestError,
    ClosedDatasetError,
    EntityAlreadyExistsError,
    EntityNotFoundError,
    ForbiddenOperationError,
    InvalidTextSearchError,
    MissingInputParamError,
    UnauthorizedError,
    WrongTaskError,
)
from argilla_server.logging import configure_logging
from argilla_server.models import User
from argilla_server.pydantic_v1 import ValidationError
from argilla_server.pydantic_v1.errors import ConfigError
from argilla_server.routes import api_router
from argilla_server.security import auth
from argilla_server.settings import settings
from argilla_server.static_rewrite import RewriteStaticFiles

_LOGGER = logging.getLogger("argilla")


def create_server_app() -> FastAPI:
    """Configure the argilla server"""

    app = FastAPI(
        title="Argilla",
        description="Argilla API",
        # Disable default openapi configuration
        openapi_url="/api/docs/spec.json",
        docs_url="/api/docs" if settings.docs_enabled else None,
        redoc_url=None,
        version=str(argilla_version),
    )

    @app.get("/docs", include_in_schema=False)
    async def redirect_docs():
        return RedirectResponse(url=f"{settings.base_url}api/docs")

    @app.get("/api", include_in_schema=False)
    async def redirect_api():
        return RedirectResponse(url=f"{settings.base_url}api/docs")

    for app_configure in [
        configure_app_logging,
        configure_database,
        configure_storage,
        configure_telemetry,
        configure_middleware,
        configure_api_exceptions,
        configure_app_security,
        configure_api_router,
        configure_app_statics,
    ]:
        app_configure(app)

    # This if-else clause is needed to simplify the test dependencies setup. Otherwise we cannot override dependencies
    # easily. We can review this once we have separate fastapi application for the api and the webapp.
    if settings.base_url and settings.base_url != "/":
        _app = FastAPI(docs_url=None, redoc_url=None, openapi_url=None)
        _app.mount(settings.base_url, app)
        return _app
    else:
        return app


def configure_middleware(app: FastAPI):
    """Configures fastapi middleware"""

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.cors_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.add_middleware(BrotliMiddleware, minimum_size=512, quality=7)


def configure_api_exceptions(api: FastAPI):
    """Configures fastapi exception handlers"""
    api.exception_handler(Exception)(APIErrorHandler.common_exception_handler)
    api.exception_handler(EntityNotFoundError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(UnauthorizedError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(ForbiddenOperationError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(EntityAlreadyExistsError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(ClosedDatasetError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(ValidationError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(AssertionError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(WrongTaskError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(MissingInputParamError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(RequestValidationError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(InvalidTextSearchError)(APIErrorHandler.common_exception_handler)
    api.exception_handler(BadRequestError)(APIErrorHandler.common_exception_handler)


def configure_api_router(app: FastAPI):
    """Configures and set the api router to app"""
    app.include_router(api_router, prefix="/api")


def configure_app_statics(app: FastAPI):
    """Configure static folder for app"""

    parent_path = Path(__file__).parent.absolute()
    statics_folder = Path(os.path.join(parent_path, "static"))
    if not (statics_folder.exists() and statics_folder.is_dir()):
        return

    def _create_statics_folder(path_from):
        """
        Application statics will be created with a parameterized baseUrl variable.

        This function will replace the variable by the real runtime value found in settings.base_url

        This allow us to deploy the argilla server under a custom base url, even when webapp does not
        support it.

        """
        BASE_URL_VAR_NAME = "@@baseUrl@@"
        temp_dir = tempfile.mkdtemp()
        new_folder = shutil.copytree(path_from, temp_dir + "/statics")
        base_url = helpers.remove_suffix(settings.base_url, suffix="/")
        for extension in ["*.js", "*.html"]:
            for file in glob.glob(
                f"{new_folder}/**/{extension}",
                recursive=True,
            ):
                helpers.replace_string_in_file(
                    file,
                    string=BASE_URL_VAR_NAME,
                    replace_by=base_url,
                )

        return new_folder

    temp_statics = _create_statics_folder(statics_folder)

    app.mount(
        "/",
        RewriteStaticFiles(directory=temp_statics, html=True, check_dir=False),
        name="static",
    )


def configure_storage(app: FastAPI):
    def _on_backoff(event):
        _LOGGER.warning(
            f"Connection to {settings.obfuscated_elasticsearch()} is not ready. "
            f"Tried {event['tries']} times. Retrying..."
        )

    @backoff.on_exception(
        lambda: backoff.constant(interval=15),
        ConfigError,
        max_time=60,
        on_backoff=_on_backoff,
    )
    def _setup_elasticsearch():
        try:
            backend = GenericElasticEngineBackend.get_instance()
            dataset_records: DatasetRecordsDAO = DatasetRecordsDAO(backend)
            datasets: DatasetsDAO = DatasetsDAO.get_instance(
                es=backend,
                records_dao=dataset_records,
            )
            datasets.init()
            dataset_records.init()
        except GenericSearchError as error:
            raise ConfigError(
                f"Your Elasticsearch endpoint at {settings.obfuscated_elasticsearch()} "
                "is not available or not responding.\n"
                "Please make sure your Elasticsearch instance is launched and correctly running and\n"
                "you have the necessary access permissions. "
                "Once you have verified this, restart the argilla server.\n"
            ) from error

    @app.on_event("startup")
    async def setup_elasticsearch():
        _setup_elasticsearch()


def configure_app_security(app: FastAPI):
    auth.configure_app(app)


def configure_app_logging(app: FastAPI):
    """Configure app logging using"""
    app.on_event("startup")(configure_logging)


def configure_telemetry(app: FastAPI):
    message = "\n"
    message += inspect.cleandoc(
        """
        Argilla uses telemetry to report anonymous usage and error information.

        You can know more about what information is reported at:

            https://docs.argilla.io/en/latest/reference/telemetry.html

        Telemetry is currently enabled. If you want to disable it, you can configure
        the environment variable before relaunching the server:
    """
    )
    message += "\n\n    "
    message += "#set ARGILLA_ENABLE_TELEMETRY=0" if os.name == "nt" else "$>export ARGILLA_ENABLE_TELEMETRY=0"
    message += "\n"

    @app.on_event("startup")
    async def check_telemetry():
        if settings.enable_telemetry:
            print(message, flush=True)


_get_db_wrapper = contextlib.asynccontextmanager(get_async_db)


def configure_database(app: FastAPI):
    def _user_has_default_credentials(user: User):
        return user.api_key == DEFAULT_API_KEY or accounts.verify_password(DEFAULT_PASSWORD, user.password_hash)

    def _log_default_user_warning():
        _LOGGER.warning(
            f"User {DEFAULT_USERNAME!r} with default credentials has been found in the database. "
            "If you are using argilla in a production environment this can be a serious security problem. "
            f"We recommend that you create a new admin user and then delete the default {DEFAULT_USERNAME!r} one."
        )

    @app.on_event("startup")
    async def log_default_user_warning_if_present():
        async with _get_db_wrapper() as db:
            default_user = await accounts.get_user_by_username(db, DEFAULT_USERNAME)
            if default_user and _user_has_default_credentials(default_user):
                _log_default_user_warning()


app = create_server_app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest import mock

import pytest
from argilla_server.enums import UserRole
from argilla_server.errors.future import AuthenticationError
from argilla_server.models import User
from argilla_server.security.authentication import JWT
from argilla_server.security.authentication.oauth2 import OAuth2Settings
from httpcore import URL
from httpx import AsyncClient
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from tests.factories import AdminFactory, AnnotatorFactory


@pytest.fixture
def disabled_oauth_settings() -> OAuth2Settings:
    return OAuth2Settings(enabled=False)


@pytest.fixture
def default_oauth_settings() -> OAuth2Settings:
    return OAuth2Settings.from_dict(
        {
            "enabled": True,
            "providers": [
                {
                    "name": "huggingface",
                    "client_id": "client_id",
                    "client_secret": "client_secret",
                }
            ],
        }
    )


@pytest.mark.asyncio
class TestOauth2:
    async def tests_list_providers_with_default_config(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.get("/api/v1/oauth2/providers", headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {"items": []}

    async def test_list_providers_with_oauth_disabled(
        self, async_client: AsyncClient, owner_auth_header: dict, disabled_oauth_settings: OAuth2Settings
    ):
        with mock.patch(
            "argilla_server.security.settings.Settings.oauth", new_callable=lambda: disabled_oauth_settings
        ):
            response = await async_client.get("/api/v1/oauth2/providers", headers=owner_auth_header)
            assert response.status_code == 200
            assert response.json() == {"items": []}

    async def test_list_provider_with_oauth_disabled_from_settings(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        default_oauth_settings.enabled = False
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get("/api/v1/oauth2/providers", headers=owner_auth_header)
            assert response.status_code == 200
            assert response.json() == {"items": []}

    async def test_list_providers(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get("/api/v1/oauth2/providers", headers=owner_auth_header)
            assert response.status_code == 200
            assert response.json() == {"items": [{"name": "huggingface"}]}

    async def test_provider_huggingface_authentication(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/authentication", headers=owner_auth_header
            )
            assert response.status_code == 303

            redirect_url = URL(response.headers.get("location"))
            assert redirect_url.scheme == b"https"
            assert redirect_url.host == b"huggingface.co"
            assert b"/oauth/authorize?response_type=code&client_id=client_id" in redirect_url.target

    async def test_provider_authentication_with_oauth_disabled(
        self,
        async_client: AsyncClient,
        owner_auth_header: dict,
        disabled_oauth_settings: OAuth2Settings,
    ):
        with mock.patch(
            "argilla_server.security.settings.Settings.oauth", new_callable=lambda: disabled_oauth_settings
        ):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/authentication", headers=owner_auth_header
            )
            assert response.status_code == 404

    async def test_provider_authentication_with_oauth_disabled_and_provider_defined(
        self,
        async_client: AsyncClient,
        owner_auth_header: dict,
        default_oauth_settings: OAuth2Settings,
    ):
        default_oauth_settings.enabled = False
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/authentication", headers=owner_auth_header
            )
            assert response.status_code == 404

    async def test_provider_authentication_with_invalid_provider(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/invalid/authentication", headers=owner_auth_header
            )
            assert response.status_code == 404

    async def test_provider_huggingface_access_token(
        self,
        async_client: AsyncClient,
        db: AsyncSession,
        owner_auth_header: dict,
        default_oauth_settings: OAuth2Settings,
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            with mock.patch(
                "argilla_server.security.authentication.oauth2.client_provider.OAuth2ClientProvider._fetch_user_data",
                return_value={"preferred_username": "username", "name": "name"},
            ):
                response = await async_client.get(
                    "/api/v1/oauth2/providers/huggingface/access-token",
                    params={"code": "code", "state": "valid"},
                    headers=owner_auth_header,
                    cookies={"oauth2_state": "valid"},
                )

                assert response.status_code == 200

                json_response = response.json()
                assert JWT.decode(json_response["access_token"])["username"] == "username"
                assert json_response["token_type"] == "bearer"

                user = (await db.execute(select(User).where(User.username == "username"))).scalar_one_or_none()
                assert user is not None
                assert user.role == UserRole.annotator

    async def test_provider_access_token_with_oauth_disabled(
        self,
        async_client: AsyncClient,
        owner_auth_header: dict,
        disabled_oauth_settings: OAuth2Settings,
    ):
        with mock.patch(
            "argilla_server.security.settings.Settings.oauth", new_callable=lambda: disabled_oauth_settings
        ):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/access-token", headers=owner_auth_header
            )
            assert response.status_code == 404

    async def test_provider_access_token_with_invalid_provider(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/invalid/authentication", headers=owner_auth_header
            )
            assert response.status_code == 404

    async def test_provider_access_token_with_not_found_code(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/access-token", headers=owner_auth_header
            )
            assert response.status_code == 400
            assert response.json() == {"detail": "'code' parameter was not found in callback request"}

    async def test_provider_access_token_with_not_found_state(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/access-token", params={"code": "code"}, headers=owner_auth_header
            )
            assert response.status_code == 400
            assert response.json() == {"detail": "'state' parameter was not found in callback request"}

    async def test_provider_access_token_with_invalid_state(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            response = await async_client.get(
                "/api/v1/oauth2/providers/huggingface/access-token",
                params={"code": "code", "state": "invalid"},
                headers=owner_auth_header,
                cookies={"oauth2_state": "valid"},
            )
            assert response.status_code == 400
            assert response.json() == {"detail": "'state' parameter does not match"}

    async def test_provider_access_token_with_authentication_error(
        self, async_client: AsyncClient, owner_auth_header: dict, default_oauth_settings: OAuth2Settings
    ):
        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            with mock.patch(
                "argilla_server.security.authentication.oauth2.client_provider.OAuth2ClientProvider._fetch_user_data",
                side_effect=AuthenticationError("error"),
            ):
                response = await async_client.get(
                    "/api/v1/oauth2/providers/huggingface/access-token",
                    params={"code": "code", "state": "valid"},
                    headers=owner_auth_header,
                    cookies={"oauth2_state": "valid"},
                )
                assert response.status_code == 401
                assert response.json() == {"detail": "error"}

    async def test_provider_access_token_with_unauthorized_user(
        self,
        async_client: AsyncClient,
        db: AsyncSession,
        owner_auth_header: dict,
        default_oauth_settings: OAuth2Settings,
    ):
        admin = await AdminFactory.create()

        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            with mock.patch(
                "argilla_server.security.authentication.oauth2.client_provider.OAuth2ClientProvider._fetch_user_data",
                return_value={"preferred_username": admin.username, "name": admin.first_name},
            ):
                response = await async_client.get(
                    "/api/v1/oauth2/providers/huggingface/access-token",
                    params={"code": "code", "state": "valid"},
                    headers=owner_auth_header,
                    cookies={"oauth2_state": "valid"},
                )
                assert response.status_code == 401
                assert response.json() == {"detail": "Could not authenticate user"}

    async def test_provider_access_token_with_same_username(
        self,
        async_client: AsyncClient,
        db: AsyncSession,
        owner_auth_header: dict,
        default_oauth_settings: OAuth2Settings,
    ):
        user = await AnnotatorFactory.create()

        with mock.patch("argilla_server.security.settings.Settings.oauth", new_callable=lambda: default_oauth_settings):
            with mock.patch(
                "argilla_server.security.authentication.oauth2.client_provider.OAuth2ClientProvider._fetch_user_data",
                return_value={"preferred_username": user.username, "name": user.first_name},
            ):
                response = await async_client.get(
                    "/api/v1/oauth2/providers/huggingface/access-token",
                    params={"code": "code", "state": "valid"},
                    headers=owner_auth_header,
                    cookies={"oauth2_state": "valid"},
                )
                # This will throw an error once we detect users created by OAuth2
                assert response.status_code == 200
